/* 
 * Copyright (C) 2013 Olivier Catoni and Thomas Mainguy
 * contact : catoni@noos.fr Thomas.Mainguy@ens.fr
 *
 * This file is part of ToricGrammars : a statistical learning model 
 * for linguistics
 *
 * ToricGrammars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ToricGrammars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "trees.h"
#include <stdio.h>

void rollover(wxChar *array, size_t &start, size_t pos, size_t &end);

void copyHist( wxChar * &hist, wxChar * const &array, size_t start,
								                size_t sEnd, size_t end); 


int main(int argc, const char *argv[] ) {

	int i,j,k;
	int labelMax;
	getSeed(_("_seed"));
	int maxOut, maxIn, iter, maxSplit, maxParse, weight, cycles, repeats;
	TreeNode grammar;
	TreeNode outputText;
	TreeNode inputText;
	TreeNode patterns;
	TreeNode reference;
	TreeNode newExp;
	bool reload;

	if (argc < 6) { 
		cout << "usage is " << argv[0] << " file maxOut maxIn number_of_cycles repeats [production_trials=10000] [number_of_innovations=10000] [number_of_parses=10000] [weight=1] \n";
		return 1;
	}

		wxString fileName(argv[1],wxConvUTF8);
		cout << "filename = " << fileName.mb_str(wxConvUTF8) << endl;
		sscanf(argv[2], "%d", &maxOut);
		sscanf(argv[3], "%d", &maxIn);
		sscanf(argv[4], "%d", &cycles); 
		sscanf(argv[5], "%d", &repeats); 
		if (argc > 6) sscanf(argv[6], "%d", &iter); else iter = 10000;
		if (argc > 7) sscanf(argv[7], "%d", &maxSplit); else maxSplit = 10000;
		if (argc > 8) sscanf(argv[8], "%d", &maxParse); else maxParse = 10000;
		if (argc > 9) sscanf(argv[9], "%d", &weight); else weight = 1;
		maxOut *= weight;
		maxIn *= weight;
		
		inputText.readTextFB(fileName, _("[0 "), weight);
		// inputText.addPatterns(patterns, _("[0 "), weight); 
		// replaces rare words by patterns. Not successful.
		patterns = inputText; // using patterns if required 
		inputText.printS(_("trainingSet.txt"));
		patterns.printS(_("patterns.txt"));

		Label label(&grammar, maxOut, maxIn);
		Label refLabel(&grammar, &reference);

		for (;repeats;repeats--) {
			labelMax = grammar.m_labelMax;
			grammar = patterns;
			grammar.m_labelMax = labelMax;
			for (i = 0, k=0; (i<maxSplit)&&label.create(); i++) {
				for (j=maxParse;j&&label.parse();j--,k++) {}
			}
			refLabel.addToRef(maxParse);
		}

		cerr << "Reference is learnt !\n";
			
		grammar = inputText;
			for (;cycles;cycles--) {
				for (j=maxParse;j&&refLabel.narrowParse();j--) {} 
				// parses the new text with the help of the reference grammar

				for (i=iter;i;i--) { // merges the grammar to produce a new text
					TreeNode outputText(grammar);
					if (outputText.mergeGrammar()) {
						grammar = outputText;
						break;
					}
				} 
				if (i == 0) { // check whether the rejection method worked 
					cout << "mergeGrammar failed !\n";  
					break;
				}
				grammar.gatherNew(_("[0 "), inputText, newExp); // gathers new sentences

			}

		grammar.printS(_("grammar.txt"));
		reference.printS(_("reference.txt"));
		newExp.printS(_("newSentences.txt"));
		saveSeed(_("_seed"));
		return 0;
}
