/*
 * Copyright (C) 2013 Olivier Catoni and Thomas Mainguy
 * contact : catoni@noos.fr Thomas.Mainguy@ens.fr
 *
 * This file is part of ToricGrammars : a statistical learning model 
 * for linguistics
 *
 * ToricGrammars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ToricGrammars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "trees.h"

char TreeNode::m_buff[BUFFLENGTH*sizeof(wxChar)];

/*
 *  * A small insert function. Should be replaced by something 
 *   * more serious, where the index is not limited to 32 bits
 *    * and is stored in an array.
 *     */
template <>
void Tree<Node>::insert(int index) {
  int sonIndex;
  m_pos = m_root;
  while(index != 1) {
    sonIndex = index & LASTBIT;
    if (!m_pos->m_Son[sonIndex]) {
      m_pos->m_Son[sonIndex] = new Node(_("wow"));
    }
    m_pos = m_pos->m_Son[sonIndex];
    index >>= 1;
  }
}

void rollover(wxChar *array, size_t &start, size_t pos, size_t &end) {
	if ((pos <=start) || (pos > end)) return; 
	assert(end + pos - start <= BUFFLENGTH);
	memcpy((void *) (array+end), (const void *) (array + start), 
									(pos - start)*sizeof(wxChar));
	end += pos - start; 
	start = pos;
}	

template <>
int Tree<Node>::growTree() {
	cout << "growTree(Tree<TNode> *)" << "\n\n";
	m_root = new Node(_("root"));
	insert(2); // binary 10 
	insert(3); // binary 11 
 	insert(4); // binary 100
	insert(6); // binary 110
	insert(5); 
	insert(7); 
	/*
	insert(123); 
	insert(40); 
	insert(78); 
	*/
	cout << "\n\n";
	return 0;
}

void saveSeed(wxString const &fileName) { 
		unsigned int seed1, seed2;
		get_seed(&seed1, &seed2);
		wxFile fileOut(fileName, wxFile::write); 
		fileOut.Write((void *) &seed1, sizeof(seed1));
		fileOut.Write((void *) &seed2, sizeof(seed2));
}

void getSeed(wxString const &fileName) { 
		unsigned int seed1, seed2;
		wxFile fileIn(fileName, wxFile::read); 
		if (!fileIn.IsOpened()) { 
						set_seed(102834, 12983); 
		} else {
			fileIn.Read((void *) &seed1, sizeof(seed1));
			fileIn.Read((void *) &seed2, sizeof(seed2));
			set_seed(seed1, seed2);
		}
}

Label::Label(TreeNode *tree, TreeNode *reference) : m_label(0), m_count(0), 
				m_stem(new wxChar[BUFFLENGTH]), m_tree(tree), m_grammar(reference) {}

Label::Label(TreeNode *tree, int maxOut, int maxIn) : m_label(0), 
				m_count(0), m_stem(new wxChar[BUFFLENGTH]), m_tree(tree), m_grammar(tree) { 
	m_max[0] = maxIn;
	m_max[1] = maxOut;
}

void Label::addToRef(int maxRemap) {
	int i;
	if (m_grammar->m_root == 0) {
		*m_grammar = *m_tree;
	} else {
		for (i = maxRemap; i&&remap();i--) {}
		m_grammar->m_root->add(m_tree->m_root);
	}
}

void Label::maxOut(int const &max) {m_max[1] = max;}  
void Label::maxIn(int const &max) {m_max[0] = max;}  

Label::~Label() {
	delete [] m_stem;
}

void TreeNode::printS(wxString const & filename) { 
	wxFileOutputStream fileStream(filename);
	wxTextOutputStream textOut(fileStream);
	printS(textOut, m_root, 0);
	textOut << _("\n\n");
}

void TreeNode::printS(wxTextOutputStream &textOut, Node * const &node, 
								const int &pos) {
	int i;
	static int count = 0;
	if (node == 0) { 
		wxChar *data = (wxChar *) m_buff; 
		textOut << count << _("\t");
		for (i=0;i<((pos>>3)/sizeof(wxChar));i++) {
			textOut << data[i];
		}
		textOut << endl;
		count = 0;
		return;
	}
	if ((node->m_count[0] + node->m_count[1]) < count) {
		wxChar *data = (wxChar *) m_buff;
		textOut << count - node->m_count[0] - node->m_count[1] << _("\t");
		for (i=0;i<((pos>>3)/sizeof(wxChar));i++) {
			textOut << data[i];
		}
		textOut << endl;
	}
	if (node->m_length.getLength()) {
		size_t bitLength;
		memcpy(m_buff+(pos>>3),node->m_data, node->m_length.getLength());
		bitLength = node->m_length.getBitLength();
		for (i=0;i<=1;i++) { 
			if (i) m_buff[(pos>>3)+(bitLength>>3)] |= (1<<(bitLength&mask7));
			else m_buff[(pos>>3)+(bitLength>>3)] &= ~(1<<(bitLength&mask7));
			if (count = node->m_count[i]) printS(textOut,node->m_Son[i],((pos>>3)<<3)+bitLength+1);
		}
	} else { 
		for (i=0;i<=1;i++) { 
			if (i) m_buff[pos>>3] |= (1<<(pos&mask7));
			else m_buff[pos>>3] &= ~(1<<(pos&mask7));
			if (count = node->m_count[i]) printS(textOut,node->m_Son[i],pos+1);
		}
	}
	count = 0;
	return;
}	

size_t TreeNode::sample_p(wxChar *array, Node *a_node, 
								size_t a_pos, size_t a_end, 
								wxString const &pattern, wxString const &exclude) {

	size_t length, bitLength, s_end, s_ptr;
	int i,j,k;

	if (a_node == 0) {
		return 0;
	}

	if (length = (a_node->m_length.getLength()-(m_offset>>3))) {
		bitLength = a_node->m_length.getBitLength() - m_offset;
		memcpy(((char *)array) + (a_pos>>3), (void *) (a_node->m_data + 
														(m_offset>>3)),
			length);
		a_pos = (a_pos & mask2) + bitLength;
	}
	m_offset = 0;

	length = pattern.length();
	s_end = ((a_pos+1)>>3)/sizeof(wxChar);

	for (i = a_node->draw(), j=2; j; i=1-i, j--) {
		if (m_Count = a_node->m_count[i]) {
			if (i) ((char *) array)[a_pos>>3] |= (1<<(a_pos&mask7));
			else ((char *) array)[a_pos>>3] &= ~(1<<(a_pos&mask7));

  		for (s_ptr= a_end; s_ptr<s_end; s_ptr++) {
				for (k=0;k< length;k++) {
					if (array[s_ptr] == pattern[k]) {
						break;	
					}
				}	
				if (k<length) {
					break;
				}
			}
			if (s_ptr<s_end) {
				if ((s_ptr == s_end - 1)&& (*array != '0')&&
							(exclude != wxString(array,s_ptr))&&
							((a_node->m_Son[i]==0)||
								(m_Count > 
							  	(a_node->m_Son[i]->m_count[0] 
								 		+ a_node->m_Son[i]->m_count[1])))) {
					return s_end;
				} else {
					return 0;
				}
			} 
			s_end = sample_p(array, a_node->m_Son[i], a_pos + 1, s_end, pattern, 
											exclude);
			if (s_end) { return s_end; }
		}
	}
	return 0;
}

size_t TreeNode::sample_p(wxChar *array, size_t start, size_t end, wxString 
								const & pattern, wxString const & exclude) {
	int count;
	size_t r_value;
	if (!(count = search(array, start, end))) {
		return end;
	}
	r_value = sample_p(array + end , m_pos, 0, 0, pattern, exclude);
	if (r_value) {
		return end + r_value;
	} else {
		return end;
	}
}

void TreeNode::merge_all(wxString const &oldLabel, wxString const &newLabel) { 
	bool output = false;
	while (merge(_("[")+oldLabel+_(" "), newLabel)) {output = true;}
	while (merge(_("]")+oldLabel+_(" "), newLabel)) {output = true;}
	if (output) m_labelCount--;
}

void Label::narrowParse(parseArg arg) {
	static size_t length, bitLength, charPtr, charPos, lastWhite, v_mid;
	int i,j;
	static parseArg argBuff;
	static wxString rightLabel;
	static wxChar * buff;

	// cerr << "narrowParse !\n";
	if (arg.aIsOk&&((arg.node==0)||(m_tree->m_Count>(arg.node->m_count[0]
																 +arg.node->m_count[1])))) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 

		if ((((m_stem[arg.mid] == '[')&&(m_stem[arg.mid+1]!='0'))
														||(m_stem[arg.mid] == ']'))&&
										(TreeNode::nextTok(m_stem,arg.mid,charPos)==charPos)) {
/* // we do not need label identification here 
			if ((rightLabel=wxString(m_stem+arg.mid+1, charPos - arg.mid - 2))!=arg.leftLabel) {
				m_tree->merge_all(rightLabel, arg.leftLabel);	
				if (m_tree != m_grammar) {
					m_grammar->merge_all(rightLabel, arg.leftLabel);
				}
				m_end = charPos; 
			}
			*/
			return;
		} else 
	// if (m_grammar->search(m_stem,arg.mid,charPos) <= m_max[1-arg.aIsOut]) 
		{
			wxString string;
/* // we do not need to compute the right label, since we do not map labels
			buff = (wxChar *) m_tree->m_buff;
			memcpy((void *) (buff + arg.mid), (void *)(m_stem + arg.mid), 
											(charPos - arg.mid)*sizeof(wxChar));
			if (arg.aIsOut) { 
				buff[charPos] = '[';
			} else {
				buff[charPos] = ']';
			}
			v_mid = m_grammar->sample_p(buff, arg.mid, charPos+1, _(" "));
			// if ((v_mid > charPos+1)&&(buff[charPos+1]!='0')) 
			if (v_mid > charPos+1) {
				rightLabel = wxString(buff+charPos+1, v_mid-charPos-2);
			} else {
				rightLabel.clear();
			}
*/
			if (arg.aIsOut) {
				string = _("]");
			} else {
				string = _("["); // should always be the case, improve code once tested
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem, arg.mid);
			m_tree->add_all(string,1);
			if (arg.aIsOut) {
				string = _("[");
			} else {
				string = _("]"); // should always be the case, improve code once tested
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem + arg.mid, charPos - arg.mid);
			m_tree->add_all(string,1);

			if (m_tree->search_g(m_stem, 0, charPos) <= 0) { 
				cerr << wxString(m_stem, charPos).mb_str(wxConvUTF8) << "|\n";
				m_tree->printS(_("grammarDebug.txt"));
				assert(false);
			}
			// assert(m_tree->search_g(m_stem, 0, charPos)>0);
			m_tree->remove_all(m_stem, 0, charPos, 1);
			m_end = charPos; // will cause all recursive calls to return

/* // no label mapping in narrow parse !
			if ((!rightLabel.empty())&&(rightLabel!=arg.leftLabel)) {
				m_grammar->merge_all(rightLabel, arg.leftLabel);
				if (m_tree != m_grammar) {
					m_tree->merge_all(rightLabel, arg.leftLabel);
				}
			}
*/

			return;
		}
		/* It seems that we could move the two previous returns here ? */
	}
	if (arg.node == 0) { 
		return;
	}
	/* Shouldn't we return also when arg.node->m_count[0] 
	 * == arg.node->m_count[1] = 0 ? 
	 * */

	if (length = arg.node->m_length.getLength()) { 
		bitLength = arg.node->m_length.getBitLength();
		memcpy(((char *) m_stem) + (arg.bitPos>>3), (void *) arg.node->m_data,
				length);
		arg.bitPos = (arg.bitPos & mask2) + bitLength;
	}

	if ((!arg.aIsOk)&&((arg.node->m_count[0]&&arg.node->m_count[1])||
													(((arg.node->m_Son[0]==0)||(arg.node->m_count[0]==0))&&
													 ((arg.node->m_Son[1]==0)||(arg.node->m_count[1]==0)))
													)) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		for ( lastWhite = charPtr = arg.mid; charPtr < charPos; charPtr++) {
			if (m_stem[charPtr] == ' ') { 
				lastWhite = charPtr;
			} 
		}
		if (lastWhite > arg.mid) { 
			for (charPtr = arg.mid; charPtr <= lastWhite; charPtr++) { 
				if (m_stem[charPtr] == '[') { 
					arg.aIsOut = 1;
					return; // abandon this branch of the tree
				} 
				if 	//(
					(m_stem[charPtr] == ' ')
// &&(m_grammar->search(m_stem, 0, charPtr+1)<=m_max[arg.aIsOut])
						//) 
				{
					if (m_grammar->search(m_stem, 0, charPtr+1)==0) {
						return; // abandon this branch of the tree 
					}

					buff = (wxChar *) m_tree->m_buff;
					memcpy((void *) buff, (void *) m_stem, (charPtr+1)*sizeof(wxChar));

/*
					if (arg.aIsOut) {
						buff[charPtr+1] = ']';
					} else {
						buff[charPtr+1] = '[';
					}
*/

					buff[charPtr+1] = '[';
					v_mid = m_grammar->sample_p(buff, 0, charPtr+2, _(" "));
					if ((v_mid > charPtr+2)&&(buff[charPtr+2]!='0')) {
						arg.leftLabel = wxString(buff+charPtr+2, v_mid-charPtr-3);
						arg.mid = charPtr+1;
						arg.aIsOk = true;
						break;
					} 

				}
			} 

		} 
	}

	for (i = arg.node->draw(), j=2; j; i=1-i, j--) { 
		if (m_tree->m_Count = arg.node->m_count[i]) { 
			if (i) ((char*)m_stem)[arg.bitPos>>3] |= (1<<(arg.bitPos&mask7));
			else ((char*)m_stem)[arg.bitPos>>3] &= ~(1<<(arg.bitPos&mask7));
			argBuff = arg;
			argBuff.bitPos++; argBuff.node = arg.node->m_Son[i];
			narrowParse(argBuff);
			if (m_end) return; // once m_en > 0, the recursive calls all return
		} 
	}
}

bool Label::narrowParse() {
	m_end = 0;
	m_tree->setCount();
	parseArg arg(m_tree->m_root);
	narrowParse(arg);
	if (m_end) { 
		return true;
	} else {
		return false;
	}
}

void Label::broadParse(parseArg arg) {
	static size_t length, bitLength, charPtr, charPos, lastWhite, v_mid;
	int i,j;
	static parseArg argBuff;
	static wxString rightLabel;
	static wxChar * buff;

	// cerr << "split !\n";
	if (arg.aIsOk&&((arg.node==0)||(m_tree->m_Count>(arg.node->m_count[0]
																 +arg.node->m_count[1])))) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 

		if ((((m_stem[arg.mid] == '[')&&(m_stem[arg.mid+1]!='0'))
														||(m_stem[arg.mid] == ']'))&&
										(TreeNode::nextTok(m_stem,arg.mid,charPos)==charPos)) {
			if ((rightLabel=wxString(m_stem+arg.mid+1, charPos - arg.mid - 2))!=arg.leftLabel) {
				m_tree->merge_all(rightLabel, arg.leftLabel);	
				cerr << "Performing label identification !\n";
				if (m_tree != m_grammar) {
					m_grammar->merge_all(rightLabel, arg.leftLabel);
				}
				m_end = charPos; 
			}
			return;
		} else if (m_grammar->search(m_stem,arg.mid,charPos) <= m_max[1-arg.aIsOut]) {
			wxString string;

			buff = (wxChar *) m_tree->m_buff;
			memcpy((void *) (buff + arg.mid), (void *)(m_stem + arg.mid), 
											(charPos - arg.mid)*sizeof(wxChar));
			if (arg.aIsOut) { 
				buff[charPos] = '[';
			} else {
				buff[charPos] = ']';
			}
			v_mid = m_grammar->sample_p(buff, arg.mid, charPos+1, _(" "));
			// if ((v_mid > charPos+1)&&(buff[charPos+1]!='0')) 
			if (v_mid > charPos+1) {
				rightLabel = wxString(buff+charPos+1, v_mid-charPos-2);
			} else {
				rightLabel.clear();
			}

			if (arg.aIsOut) {
				string = _("]");
			} else {
				string = _("[");
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem, arg.mid);
			m_tree->add_all(string,1);
			if (arg.aIsOut) {
				string = _("[");
			} else {
				string = _("]");
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem + arg.mid, charPos - arg.mid);
			m_tree->add_all(string,1);

			if (m_tree->search_g(m_stem, 0, charPos) <= 0) { 
				cerr << wxString(m_stem, charPos).mb_str(wxConvUTF8) << "|\n";
				m_tree->printS(_("grammarDebug.txt"));
				assert(false);
			}
			// assert(m_tree->search_g(m_stem, 0, charPos)>0);
			m_tree->remove_all(m_stem, 0, charPos, 1);
			m_end = charPos;

			if ((!rightLabel.empty())&&(rightLabel!=arg.leftLabel)) {
				m_grammar->merge_all(rightLabel, arg.leftLabel);
				cerr << "Performing label identification !\n";
				if (m_tree != m_grammar) {
					m_tree->merge_all(rightLabel, arg.leftLabel);
				}
			}

			return;
		}
		/* It seems that we could move the two previous returns here ? */
	}
	if (arg.node == 0) { 
		return;
	}
	/* Shouldn't we return also when arg.node->m_count[0] 
	 * == arg.node->m_count[1] = 0 ? 
	 * */

	if (length = arg.node->m_length.getLength()) { 
		bitLength = arg.node->m_length.getBitLength();
		memcpy(((char *) m_stem) + (arg.bitPos>>3), (void *) arg.node->m_data,
				length);
		arg.bitPos = (arg.bitPos & mask2) + bitLength;
	}

	if ((!arg.aIsOk)&&((arg.node->m_count[0]&&arg.node->m_count[1])||
													(((arg.node->m_Son[0]==0)||(arg.node->m_count[0]==0))&&
													 ((arg.node->m_Son[1]==0)||(arg.node->m_count[1]==0)))
													)) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		for ( lastWhite = charPtr = arg.mid; charPtr < charPos; charPtr++) {
			if (m_stem[charPtr] == ' ') { 
				lastWhite = charPtr;
			} 
		}
		if (lastWhite > arg.mid) { 
			for (charPtr = arg.mid; charPtr <= lastWhite; charPtr++) { 
				if (m_stem[charPtr] == '[') { 
					arg.aIsOut = 1;
				} 
				if ((m_stem[charPtr] == ' ')
												&&(m_grammar->search(m_stem, 0, charPtr+1)<=m_max[arg.aIsOut])
												) {

					buff = (wxChar *) m_tree->m_buff;
					memcpy((void *) buff, (void *) m_stem, (charPtr+1)*sizeof(wxChar));
					if (arg.aIsOut) {
						buff[charPtr+1] = ']';
					} else {
						buff[charPtr+1] = '[';
					}
					v_mid = m_grammar->sample_p(buff, 0, charPtr+2, _(" "));
					if ((v_mid > charPtr+2)&&(buff[charPtr+2]!='0')) {
						arg.leftLabel = wxString(buff+charPtr+2, v_mid-charPtr-3);
						arg.mid = charPtr+1;
						arg.aIsOk = true;
						break;
					} 

				}
			} 

		} 
	}

	for (i = arg.node->draw(), j=2; j; i=1-i, j--) { 
		if (m_tree->m_Count = arg.node->m_count[i]) { 
			if (i) ((char*)m_stem)[arg.bitPos>>3] |= (1<<(arg.bitPos&mask7));
			else ((char*)m_stem)[arg.bitPos>>3] &= ~(1<<(arg.bitPos&mask7));
			argBuff = arg;
			argBuff.bitPos++; argBuff.node = arg.node->m_Son[i];
			parse(argBuff);
			if (m_end) return;
		} 
	}
}

void Label::parse(parseArg arg) {
	static size_t length, bitLength, charPtr, charPos, lastWhite, v_mid;
	int i,j;
	static parseArg argBuff;
	static wxString rightLabel;
	static wxChar * buff;

	// cerr << "split !\n";
	if (arg.aIsOk&&((arg.node==0)||(m_tree->m_Count>(arg.node->m_count[0]
																 +arg.node->m_count[1])))) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		if (arg.mid >= (charPos-1)) { 
			return;
		}

		if ((((m_stem[arg.mid] == '[')&&(m_stem[arg.mid+1]!='0'))
														||(m_stem[arg.mid] == ']'))&&
										(TreeNode::nextTok(m_stem,arg.mid,charPos)==charPos)) {
			rightLabel = wxString(m_stem+arg.mid+1, charPos - arg.mid - 2);

			if (rightLabel == arg.leftLabel) { // search for other labels for a 
				buff = (wxChar *) m_tree->m_buff;
				memcpy((void *) buff, (void *) m_stem, (arg.mid + 1)*sizeof(wxChar));
				v_mid = m_grammar->sample_p(buff,0,arg.mid+1,_(" "),rightLabel);
				// sample_p is called with exclude_pattern = rightlabel
				if (v_mid > (arg.mid+1)) {
					arg.leftLabel = wxString(buff+arg.mid+1, v_mid - arg.mid - 2);
					cerr << "Second label found !\n";
					// cerr << "rightLabel = " << rightLabel.mb_str(wxConvUTF8) << "|\n";
					// cerr << "leftLabel = " << arg.leftLabel.mb_str(wxConvUTF8) << "|\n";
				}
			}

			if (rightLabel!=arg.leftLabel) {
				m_tree->merge_all(rightLabel, arg.leftLabel);	
				cerr << "Performing label identification !\n";
				// cerr << "rightLabel = " << rightLabel.mb_str(wxConvUTF8) << "|\n";
				// cerr << "leftLabel = " << arg.leftLabel.mb_str(wxConvUTF8) << "|\n";
				if (m_tree != m_grammar) {
					m_grammar->merge_all(rightLabel, arg.leftLabel);
				}
				m_end = charPos; 
				// m_end = v_mid; 
			}

			return;
		} else if (m_grammar->search(m_stem,arg.mid,charPos) <= m_max[1-arg.aIsOut]) {
			wxString string;

			buff = (wxChar *) m_tree->m_buff;
			memcpy((void *) (buff + arg.mid), (void *)(m_stem + arg.mid), 
											(charPos - arg.mid)*sizeof(wxChar));
			if (arg.aIsOut) { 
				buff[charPos] = '[';
			} else {
				buff[charPos] = ']';
			}
			v_mid = m_grammar->sample_p(buff, arg.mid, charPos+1, _(" "),
											arg.leftLabel);
			// if ((v_mid > charPos+1)&&(buff[charPos+1]!='0')) 
			if (v_mid > charPos+1) {
				rightLabel = wxString(buff+charPos+1, v_mid-charPos-2);
			} else {
				rightLabel.clear();
			}

			if (arg.aIsOut) {
				string = _("]");
			} else {
				string = _("[");
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem, arg.mid);
			m_tree->add_all(string,1);
			if (arg.aIsOut) {
				string = _("[");
			} else {
				string = _("]");
			}
			string << arg.leftLabel << _(" ") << wxString(m_stem + arg.mid, charPos - arg.mid);
			m_tree->add_all(string,1);

			if (m_tree->search_g(m_stem, 0, charPos) <= 0) { 
				cerr << wxString(m_stem, charPos).mb_str(wxConvUTF8) << "|\n";
				m_tree->printS(_("grammarDebug.txt"));
				assert(false);
			}
			// assert(m_tree->search_g(m_stem, 0, charPos)>0);
			m_tree->remove_all(m_stem, 0, charPos, 1);
			m_end = charPos;

			if ((!rightLabel.empty())&&(rightLabel!=arg.leftLabel)) {
				m_grammar->merge_all(rightLabel, arg.leftLabel);
				cerr << "Performing label identification !\n";
				// cerr << "rightLabel = " << rightLabel.mb_str(wxConvUTF8) << "|\n";
				// cerr << "leftLabel = " << arg.leftLabel.mb_str(wxConvUTF8) << "|\n";
				if (m_tree != m_grammar) {
					m_tree->merge_all(rightLabel, arg.leftLabel);
				}
			}

			return;
		}
		/* It seems that we could move the two previous returns here ? */
	}
	if (arg.node == 0) { 
		return;
	}
	/* Shouldn't we return also when arg.node->m_count[0] 
	 * == arg.node->m_count[1] = 0 ? 
	 * */

	if (length = arg.node->m_length.getLength()) { 
		bitLength = arg.node->m_length.getBitLength();
		memcpy(((char *) m_stem) + (arg.bitPos>>3), (void *) arg.node->m_data,
				length);
		arg.bitPos = (arg.bitPos & mask2) + bitLength;
	}

	if ((!arg.aIsOk)&&((arg.node->m_count[0]&&arg.node->m_count[1])||
													(((arg.node->m_Son[0]==0)||(arg.node->m_count[0]==0))&&
													 ((arg.node->m_Son[1]==0)||(arg.node->m_count[1]==0)))
													)) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		for ( lastWhite = charPtr = arg.mid; charPtr < charPos; charPtr++) {
			if (m_stem[charPtr] == ' ') { 
				lastWhite = charPtr;
			} 
		}
		if (lastWhite > arg.mid) { 
			for (charPtr = arg.mid; charPtr <= lastWhite; charPtr++) { 
				if (m_stem[charPtr] == '[') { 
					arg.aIsOut = 1;
				} 
				if ((m_stem[charPtr] == ' ')
												&&(m_grammar->search(m_stem, 0, charPtr+1)<=m_max[arg.aIsOut])
												) {

					buff = (wxChar *) m_tree->m_buff;
					memcpy((void *) buff, (void *) m_stem, (charPtr+1)*sizeof(wxChar));
					if (arg.aIsOut) {
						buff[charPtr+1] = ']';
					} else {
						buff[charPtr+1] = '[';
					}
					v_mid = m_grammar->sample_p(buff, 0, charPtr+2, _(" "));
					if ((v_mid > charPtr+2)&&(buff[charPtr+2]!='0')) {
						arg.leftLabel = wxString(buff+charPtr+2, v_mid-charPtr-3);
						arg.mid = charPtr+1;
						arg.aIsOk = true;
						break;
					} 

				}
			} 

		} 
	}

	for (i = arg.node->draw(), j=2; j; i=1-i, j--) { 
		if (m_tree->m_Count = arg.node->m_count[i]) { 
			if (i) ((char*)m_stem)[arg.bitPos>>3] |= (1<<(arg.bitPos&mask7));
			else ((char*)m_stem)[arg.bitPos>>3] &= ~(1<<(arg.bitPos&mask7));
			argBuff = arg;
			argBuff.bitPos++; argBuff.node = arg.node->m_Son[i];
			parse(argBuff);
			if (m_end) return;
		} 
	}
}

bool Label::parse() {
	m_end = 0;
	m_tree->setCount();
	parseArg arg(m_tree->m_root);
	parse(arg);
	if (m_end) { 
		return true;
	} else {
		return false;
	}
} 

void Label::create(splitArg arg) {
	static size_t length, bitLength, charPtr, charPos, lastWhite, v_mid;
	int i,j;
	static splitArg argBuff;
	static wxString leftLabel, rightLabel;
	static wxChar * buff;

	// cerr << "split !\n";
	// if we have reached the end of an expression
	if (arg.aIsOk&&((arg.node==0)||(m_tree->m_Count>(arg.node->m_count[0]
																 +arg.node->m_count[1])))) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 

		memcpy((void *)(m_stem+charPos), (void *) m_stem, charPos*sizeof(wxChar));
		charPtr = charPos;
		v_mid = charPos + arg.mid;
		for (i = j = m_grammar->search(m_stem, charPtr, v_mid);
		    (j >= i)&&(charPtr > arg.mid); j = m_grammar->search(m_stem, charPtr, v_mid)) {
			charPtr = TreeNode::prevTok(m_stem, charPtr, 0);  
		}
		charPtr = TreeNode::nextTok(m_stem, charPtr, charPos);
		if ((j = m_grammar->search(m_stem, charPtr, v_mid)) != i) {
			cerr << "i = " << m_grammar->search(m_stem, charPos, v_mid) << ", j = " << j << endl;
			cerr << wxString(m_stem+charPos, v_mid - charPos).mb_str(wxConvUTF8) << "|\n";
			cerr << wxString(m_stem+charPtr, v_mid - charPtr).mb_str(wxConvUTF8) << "|\n";
			m_grammar->printS(_("grammarDebug.txt"));
			m_grammar->print(_("treeDebug.txt"));
			assert(false);
		}
		// here we use the available variable lastWhite as an index
		// unrelated to its name. 
		// a cleaner way would have been to declare 
		// size_t &index = lastWhite;
		// and use index instead of lastWhite
		for (lastWhite = charPtr; lastWhite<charPos; lastWhite++) {
			if (m_stem[lastWhite] == '[') {
				arg.aIsOut = 1;
			}
		}
		charPos += charPtr;

		if ((((m_stem[v_mid]!='[')&&(m_stem[v_mid]!=']'))||
						(TreeNode::nextTok(m_stem,v_mid,charPos) < charPos))
								&&(m_grammar->search(m_stem,v_mid,charPos) <= m_max[1-arg.aIsOut])) {
			wxString string;

			arg.mid = v_mid;

			buff = (wxChar *) m_tree->m_buff;

			memcpy((void *) buff, (void *) m_stem, arg.mid*sizeof(wxChar));
			if (arg.aIsOut) {
				buff[arg.mid] = ']';
			} else {
				buff[arg.mid] = '[';
			}
			v_mid = m_grammar->sample_p(buff, charPtr, arg.mid+1, _(" "));
			// cerr << "sample_p(" << wxString(buff, arg.mid+1).mb_str(wxConvUTF8) << ") = " 
			//			<< wxString(buff, v_mid).mb_str(wxConvUTF8) << endl; 
			if (v_mid > arg.mid+1) {
				leftLabel = wxString(buff+arg.mid+1, v_mid-arg.mid-2);
			} else {
				leftLabel.clear();
			}

			memcpy((void *) (buff + arg.mid), (void *)(m_stem + arg.mid), 
											(charPos - arg.mid)*sizeof(wxChar));
			if (arg.aIsOut) { 
				buff[charPos] = '[';
			} else {
				buff[charPos] = ']';
			}
			v_mid = m_grammar->sample_p(buff, arg.mid, charPos+1, _(" "));
			// cerr << "sample_p(" << wxString(buff+arg.mid, charPos+1-arg.mid).mb_str(wxConvUTF8) << ") = " 
			//					<< wxString(buff + arg.mid, v_mid - arg.mid).mb_str(wxConvUTF8) << endl; 
			if (v_mid > charPos+1) {
				rightLabel = wxString(buff+charPos+1, v_mid-charPos-2);
			} else {
				rightLabel.clear();
			}

			if (leftLabel.empty()) { 
				if (!rightLabel.empty()) {
					leftLabel = rightLabel; 
				} else { 
					m_tree->m_labelMax++;
					m_tree->m_labelCount++;
					if (m_grammar != m_tree) {
						m_grammar->m_labelMax++;
					}
					leftLabel << m_tree->m_labelMax;
				}
			} /* else { // leftLabel.empty() == false

				if ((!rightLabel.empty())&&(rightLabel!=leftLabel)) {
					m_grammar->merge_all(rightLabel, leftLabel);
					if (m_tree != m_grammar) {
						m_tree->merge_all(rightLabel, leftLabel);
					}
				}

			} */ 

			if (arg.aIsOut) {
				string = _("]");
			} else {
				string = _("[");
			}
			string << leftLabel << _(" ") << wxString(m_stem + charPtr, arg.mid - charPtr);
			m_tree->add_all(string,1);

			// cerr <<	"split : " << wxString(m_stem, charPos).mb_str(wxConvUTF8) << " ->\n"
			//			<< string.mb_str(wxConvUTF8) << " + ";

			if (arg.aIsOut) {
				string = _("[");
			} else {
				string = _("]");
			}
			string << leftLabel << _(" ") << wxString(m_stem + arg.mid, charPos - arg.mid);
			m_tree->add_all(string,1);

			// cerr << string.mb_str(wxConvUTF8) << endl;

			if (m_tree->search_g(m_stem, charPtr, charPos) <= 0) {
				cerr << wxString(m_stem+charPtr, charPos - charPtr).mb_str(wxConvUTF8) << "|\n";
				assert(false);
			}
			m_tree->remove_all(m_stem, charPtr, charPos, 1);
			m_end = charPos;

				if ((!rightLabel.empty())&&(rightLabel!=leftLabel)) {
					m_grammar->merge_all(rightLabel, leftLabel);
					cerr << "Performing label identification !\n";
					if (m_tree != m_grammar) {
						m_tree->merge_all(rightLabel, leftLabel);
					}
				}

			return;
		}
	}
	if (arg.node == 0) {
		return;
	}

	if (length = arg.node->m_length.getLength()) { 
		bitLength = arg.node->m_length.getBitLength();
		memcpy(((char *) m_stem) + (arg.bitPos>>3), (void *) arg.node->m_data,
				length);
		arg.bitPos = (arg.bitPos & mask2) + bitLength;
	}

	if ((!arg.aIsOk)&&arg.node->m_count[0]&&arg.node->m_count[1]) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		for ( lastWhite = charPtr = arg.mid; charPtr < charPos; charPtr++) {
			if (m_stem[charPtr] == ' ') { 
				lastWhite = charPtr;
			} 
		}
		if (lastWhite > arg.mid) { 
			for (charPtr = arg.mid; charPtr <= lastWhite; charPtr++) { 
				if (m_stem[charPtr] == '[') { 
					arg.aIsOut = 1;
				} 
			} 
			arg.mid = charPtr; // charPtr == lastWhite + 1 <= charPos
			if ((((m_stem[0] != '[')&&(m_stem[0] != ']'))||
											(TreeNode::nextTok(m_stem,0,arg.mid) < arg.mid)) 
						&&(m_grammar->search(m_stem, 0, arg.mid) <= m_max[arg.aIsOut])) {
				arg.aIsOk = true;
			}
		} 
	}

	for (i = arg.node->draw(), j=2; j; i=1-i, j--) { 
		if (m_tree->m_Count = arg.node->m_count[i]) { 
			if (i) ((char*)m_stem)[arg.bitPos>>3] |= (1<<(arg.bitPos&mask7));
			else ((char*)m_stem)[arg.bitPos>>3] &= ~(1<<(arg.bitPos&mask7));
			argBuff = arg;
			argBuff.bitPos++; argBuff.node = arg.node->m_Son[i];
			split(argBuff);
			if (m_end) return;
		} 
	}
}

void Label::split(splitArg arg) {
	static size_t length, bitLength, charPtr, charPos, lastWhite, v_mid;
	int i,j;
	static splitArg argBuff;
	static wxString leftLabel, rightLabel;
	static wxChar * buff;

	// cerr << "split !\n";
	// if we have reached the end of an expression
	if (arg.aIsOk&&((arg.node==0)||(m_tree->m_Count>(arg.node->m_count[0]
																 +arg.node->m_count[1])))) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 

		memcpy((void *)(m_stem+charPos), (void *) m_stem, charPos*sizeof(wxChar));
		charPtr = charPos;
		v_mid = charPos + arg.mid;
		for (i = j = m_grammar->search(m_stem, charPtr, v_mid);
		    (j >= i)&&(charPtr > arg.mid); j = m_grammar->search(m_stem, charPtr, v_mid)) {
			charPtr = TreeNode::prevTok(m_stem, charPtr, 0);  
		}
		charPtr = TreeNode::nextTok(m_stem, charPtr, charPos);
		if ((j = m_grammar->search(m_stem, charPtr, v_mid)) != i) {
			cerr << "i = " << m_grammar->search(m_stem, charPos, v_mid) << ", j = " << j << endl;
			cerr << wxString(m_stem+charPos, v_mid - charPos).mb_str(wxConvUTF8) << "|\n";
			cerr << wxString(m_stem+charPtr, v_mid - charPtr).mb_str(wxConvUTF8) << "|\n";
			m_grammar->printS(_("grammarDebug.txt"));
			m_grammar->print(_("treeDebug.txt"));
			assert(false);
		}
		// here we use the available variable lastWhite as an index
		// unrelated to its name. 
		// a cleaner way would have been to declare 
		// size_t &index = lastWhite;
		// and use index instead of lastWhite
		for (lastWhite = charPtr; lastWhite<charPos; lastWhite++) {
			if (m_stem[lastWhite] == '[') {
				arg.aIsOut = 1;
			}
		}
		charPos += charPtr;

		if ((((m_stem[v_mid]!='[')&&(m_stem[v_mid]!=']'))||
						(TreeNode::nextTok(m_stem,v_mid,charPos) < charPos))
								&&(m_grammar->search(m_stem,v_mid,charPos) <= m_max[1-arg.aIsOut])) {
			wxString string;

			arg.mid = v_mid;

			buff = (wxChar *) m_tree->m_buff;

			memcpy((void *) buff, (void *) m_stem, arg.mid*sizeof(wxChar));
			if (arg.aIsOut) {
				buff[arg.mid] = ']';
			} else {
				buff[arg.mid] = '[';
			}
			v_mid = m_grammar->sample_p(buff, charPtr, arg.mid+1, _(" "));
			// cerr << "sample_p(" << wxString(buff, arg.mid+1).mb_str(wxConvUTF8) << ") = " 
			//			<< wxString(buff, v_mid).mb_str(wxConvUTF8) << endl; 
			if (v_mid > arg.mid+1) {
				leftLabel = wxString(buff+arg.mid+1, v_mid-arg.mid-2);
			} else {
				leftLabel.clear();
			}

			memcpy((void *) (buff + arg.mid), (void *)(m_stem + arg.mid), 
											(charPos - arg.mid)*sizeof(wxChar));
			if (arg.aIsOut) { 
				buff[charPos] = '[';
			} else {
				buff[charPos] = ']';
			}
			v_mid = m_grammar->sample_p(buff, arg.mid, charPos+1, _(" "));
			// cerr << "sample_p(" << wxString(buff+arg.mid, charPos+1-arg.mid).mb_str(wxConvUTF8) << ") = " 
			//					<< wxString(buff + arg.mid, v_mid - arg.mid).mb_str(wxConvUTF8) << endl; 
			if (v_mid > charPos+1) {
				rightLabel = wxString(buff+charPos+1, v_mid-charPos-2);
			} else {
				rightLabel.clear();
			}

			if (leftLabel.empty()) { 
				if (!rightLabel.empty()) {
					leftLabel = rightLabel; 
				} else { 
					m_tree->m_labelMax++;
					m_tree->m_labelCount++;
					if (m_grammar != m_tree) {
						m_grammar->m_labelMax++;
					}
					leftLabel << m_tree->m_labelMax;
				}
			} /* else { // leftLabel.empty() == false

				if ((!rightLabel.empty())&&(rightLabel!=leftLabel)) {
					m_grammar->merge_all(rightLabel, leftLabel);
					if (m_tree != m_grammar) {
						m_tree->merge_all(rightLabel, leftLabel);
					}
				}

			} */ 

			if (arg.aIsOut) {
				string = _("]");
			} else {
				string = _("[");
			}
			string << leftLabel << _(" ") << wxString(m_stem + charPtr, arg.mid - charPtr);
			m_tree->add_all(string,1);

			// cerr <<	"split : " << wxString(m_stem, charPos).mb_str(wxConvUTF8) << " ->\n"
			//			<< string.mb_str(wxConvUTF8) << " + ";

			if (arg.aIsOut) {
				string = _("[");
			} else {
				string = _("]");
			}
			string << leftLabel << _(" ") << wxString(m_stem + arg.mid, charPos - arg.mid);
			m_tree->add_all(string,1);

			// cerr << string.mb_str(wxConvUTF8) << endl;

			if (m_tree->search_g(m_stem, charPtr, charPos) <= 0) {
				cerr << wxString(m_stem+charPtr, charPos - charPtr).mb_str(wxConvUTF8) << "|\n";
				assert(false);
			}
			m_tree->remove_all(m_stem, charPtr, charPos, 1);
			m_end = charPos;

				if ((!rightLabel.empty())&&(rightLabel!=leftLabel)) {
					m_grammar->merge_all(rightLabel, leftLabel);
					if (m_tree != m_grammar) {
						m_tree->merge_all(rightLabel, leftLabel);
					}
				}

			return;
		}
	}
	if (arg.node == 0) {
		return;
	}

	if (length = arg.node->m_length.getLength()) { 
		bitLength = arg.node->m_length.getBitLength();
		memcpy(((char *) m_stem) + (arg.bitPos>>3), (void *) arg.node->m_data,
				length);
		arg.bitPos = (arg.bitPos & mask2) + bitLength;
	}

	if ((!arg.aIsOk)&&arg.node->m_count[0]&&arg.node->m_count[1]) {
		charPos = (arg.bitPos >> 3)/(sizeof(wxChar)); 
		for ( lastWhite = charPtr = arg.mid; charPtr < charPos; charPtr++) {
			if (m_stem[charPtr] == ' ') { 
				lastWhite = charPtr;
			} 
		}
		if (lastWhite > arg.mid) { 
			for (charPtr = arg.mid; charPtr <= lastWhite; charPtr++) { 
				if (m_stem[charPtr] == '[') { 
					arg.aIsOut = 1;
				} 
			} 
			arg.mid = charPtr; // charPtr == lastWhite + 1 <= charPos
			if ((((m_stem[0] != '[')&&(m_stem[0] != ']'))||
											(TreeNode::nextTok(m_stem,0,arg.mid) < arg.mid)) 
						&&(m_grammar->search(m_stem, 0, arg.mid) <= m_max[arg.aIsOut])) {
				arg.aIsOk = true;
			}
		} 
	}

	for (i = arg.node->draw(), j=2; j; i=1-i, j--) { 
		if (m_tree->m_Count = arg.node->m_count[i]) { 
			if (i) ((char*)m_stem)[arg.bitPos>>3] |= (1<<(arg.bitPos&mask7));
			else ((char*)m_stem)[arg.bitPos>>3] &= ~(1<<(arg.bitPos&mask7));
			argBuff = arg;
			argBuff.bitPos++; argBuff.node = arg.node->m_Son[i];
			split(argBuff);
			if (m_end) return;
		} 
	}
}
 
void TreeNode::setCount() {
	m_Count = m_root->m_count[0]+m_root->m_count[1];
}

bool Label::create() {
	m_end = 0;
	m_tree->setCount();
	splitArg arg(m_tree->m_root);
	split(arg);
	if (m_end) { 
		return true;
	} else {
		return false;
	}
}

bool Label::split() {
	m_end = 0;
	m_tree->setCount();
	splitArg arg(m_tree->m_root);
	split(arg);
	if (m_end) { 
		return true;
	} else {
		return false;
	}
} 

bool Label::getLabel() { 
	m_end = 0;
	m_label++;
	m_count++;
	getLabel(m_tree->m_root,0,0,0); 
	if (m_end==0) return false;
	// cout << "m_stem = " << wxString(m_stem, m_end).mb_str(wxConvUTF8) << endl;
/*
	m_isOut = 1 - m_isOut;
	spread(m_stem, 0, m_end);
	m_isOut = 1 - m_isOut;
*/
	return true;
}

void Label::getLabel(Node *node, size_t a_pos, size_t a_start, int a_isOut) {
	size_t new_end = 0;
	size_t new_start = 0;
	int i;

	if (node == 0) {
		m_end = 0;
		return;
	}

	// m_count = node->m_count[0]+node->m_count[1];

	if (node->m_length.getLength()) {
		size_t bitLength = node->m_length.getBitLength();
		memcpy(((char *) m_stem)+(a_pos>>3), (void *) node->m_data, 
										node->m_length.getLength());
		a_pos = (a_pos&mask2) + bitLength;
	}	

	new_start = a_start;
	new_end = a_pos / (sizeof(wxChar)<<3);

	// if (m_end == 0) {}
	if (node->m_count[0]&&node->m_count[1]) {
	new_start = TreeNode::nextTok(m_stem, a_start, new_end, whiteSpace);
	while (a_start < new_end) {
		if (a_start < new_start) { 
			for (i = a_start;i<new_start;i++) {
				if (m_stem[i] == '[') a_isOut = 1;
			}
		}
		if (  // (m_end == 0)&&
					new_start 
					&&(m_stem[new_start-1] == ' ') 	
					// &&(m_count <= m_max[a_isOut]) 
					&&(
							(TreeNode::nextTok(m_stem,0,new_start,whiteSpace) < new_start)
							||(
								(m_stem[0]!='[') && (m_stem[0]!=']')
							)
					)	) {
			m_isOut = 1 - a_isOut;
			m_end = new_start;
			if ((m_tree->search(m_stem, 0, m_end) <= m_max[a_isOut])
											&&spread(m_stem, 0, m_end)) {
				return;
			}	
		}
		a_start = new_start; 
		new_start= TreeNode::nextTok(m_stem, new_start, new_end, whiteSpace);
	} 
	}

	/*
	if (m_end) { 
		if (node->m_count[0]&&node->m_count[1]) {
			m_isOut = 1 - m_isOut;
			if (spread(m_stem, 0, m_end)) { 
				m_isOut = 1 - m_isOut;
				return;
			} else {
				m_isOut = 1 - m_isOut;
				m_end = 0;
				// return;
			}
		}
		// if (!(node->m_count[0]||node->m_count[1])) { m_end = 0; return; }
	}
	*/

		i = node->draw(); // we randomize the choice of the new label.
		if (node->m_count[i] > 0) {
			if (i) ((char *)m_stem)[a_pos>>3] |= (1<<(a_pos&mask7));
			else ((char *)m_stem)[a_pos>>3] &= ~(1<<(a_pos&mask7));
			getLabel(node->m_Son[i],a_pos+1,new_start, a_isOut);
			if (m_end) return;
		}
		i = 1 - i;
		if ((m_end == 0)&&(node->m_count[i] > 0)) { 
			if (i) ((char *)m_stem)[a_pos>>3] |= (1<<(a_pos&mask7));
			else ((char *)m_stem)[a_pos>>3] &= ~(1<<(a_pos&mask7));
			getLabel(node->m_Son[i],a_pos+1,new_start, a_isOut);
		}

	return;
}	

bool TreeNode::merge(wxString const &string1, wxString const &newLabel) { 

	size_t length1 = string1.length();
	wxString stem2 = read(string1);
	wxChar const *array2 = stem2.c_str();
	size_t sEnd2 = stem2.length();
	size_t start2, end2, bufEnd;
	int count;
	bool next = false;

	wxString starter1;
	starter1 << string1[0] << newLabel << _(" ");

	start2 = 0; 
	end2 = TreeNode::first_of(array2, start2, sEnd2, _("\n"));
	while (start2 < sEnd2) { 

		wxString string2(array2+start2, end2 - start2);
			memcpy(m_buff,
				(void *)(string1+string2).c_str(),
						(length1+end2-start2)*sizeof(wxChar));
			bufEnd = length1 + end2-start2;
			if (count = search_g((wxChar *) m_buff, 0, bufEnd)) {
			remove_all((wxChar *) m_buff, 0, bufEnd); 

			memcpy(m_buff, 
				(const void *)(starter1+string2).c_str(),
						(starter1.length()+end2-start2)*sizeof(wxChar));
			bufEnd = starter1.length() + end2-start2;
			add_all((wxChar *) m_buff, 0, bufEnd, count); 
			next = true;
			}

		start2 = end2+1;
		end2 = TreeNode::first_of(array2, start2, sEnd2, _("\n"));
	} 
	return next;
}

bool Label::spread(wxChar const *array1, size_t start1, size_t end1) { 
	wxString stem2 = m_tree->read(array1, start1, end1);
	// cout << "stem2 = \n" << stem2.mb_str(wxConvUTF8) << endl;
	wxChar const *array2 = stem2.c_str();
	size_t sEnd2 = stem2.length();
	size_t start2, end2, bufEnd;
	int count;
	bool success = false;
	wxString newLabel;
	newLabel	<< m_label;


	wxString starter1;
	if (m_isOut) starter1 << _("[") << m_label << _(" ");
	else starter1 << _("]") << m_label << _(" ");

	wxString starter0;
	starter0 << starter1[0] << _("0 ");

	wxString starter2;
	if (m_isOut) starter2 << _("]") << m_label << _(" ");
	else starter2 << _("[") << m_label << _(" ");
  
	start2 = 0; 
	end2 = TreeNode::first_of(array2, start2, sEnd2, _("\n"));
	while (start2 < sEnd2) { 
		wxString string2(array2+start2, end2 - start2);

		if ((string2 != starter1)
				&&(m_tree->search(array2,start2,end2)<=m_max[m_isOut])) {

		if ((string2[0] == starter1[0])&&(string2 != starter0) 
			   &&(TreeNode::nextTok(array2,start2,end2) == end2)) {
			while (m_tree->merge(string2, newLabel)) {}
			string2[0] = starter2[0];
			while(m_tree->merge(string2, newLabel)) {} 
			string2[0] = starter1[0];
			m_count--;
			success = true;
		} else { 

			memcpy(m_tree->m_buff,
				(void *)(wxString(array1+start1, end1-start1)
					+wxString(array2+start2, end2-start2)).c_str(),
						(end1-start1+end2-start2)*sizeof(wxChar));
			bufEnd = end1 - start1 + end2-start2;
			if (count = m_tree->search_g((wxChar *) m_tree->m_buff, 0, bufEnd)) {
			m_tree->remove_all((wxChar *) m_tree->m_buff, 0, bufEnd); 

			memcpy(m_tree->m_buff, 
				(const void *)(starter1
					+wxString(array1+start1, end1-start1)).c_str(),
						(starter1.length()+end1-start1)*sizeof(wxChar));
			bufEnd = starter1.length() + end1-start1;
			m_tree->add_all((wxChar *) m_tree->m_buff, 0, bufEnd, count); 

			memcpy(m_tree->m_buff, 
				(const void *)(starter2
					+wxString(array2+start2, end2-start2)).c_str(),
						(starter2.length()+end2-start2)*sizeof(wxChar));
			bufEnd = starter2.length() + end2-start2;
			m_tree->add_all((wxChar *) m_tree->m_buff, 0, bufEnd, count); 

			success = true;
			}
		}
		}
		start2 = end2+1;
		end2 = TreeNode::first_of(array2, start2, sEnd2, _("\n"));
	} 
	return success;
}

bool Label::spreadLabel() { 

	bool next; 
	bool modified = false;

	do {
	m_isOut = 1 - m_isOut;
	next = false;
	wxString starter1;

	if (m_isOut) starter1 << _("[") << m_label << _(" ");
	else starter1 << _("]") << m_label << _(" ");
	// cout << "starter1 = \n" << starter1.mb_str(wxConvUTF8) << endl;

	wxString stem1 = m_tree->read(starter1);
	// cout << "stem1 = \n" << stem1.mb_str(wxConvUTF8) << endl;

	size_t start1, end1, sEnd1;

	wxChar const *array1 = stem1.c_str();
	sEnd1 = stem1.length();
	
	for (
		start1 = TreeNode::first_not_of(array1, 0, sEnd1, _("\n")),
			end1 = TreeNode::first_of(array1, start1, sEnd1, _("\n"));
		start1 < end1; 
		start1 = TreeNode::first_not_of(array1, end1, sEnd1, _("\n")),
		end1 = TreeNode::first_of(array1, start1, sEnd1, _("\n"))) {

		if (m_tree->search(array1, start1, end1) > m_max[1-m_isOut]) continue; 
		if (spread(array1,start1,end1)) { next = true; modified = true;}

	}

	} while(next);
	return modified;
}

Node::Node(Length const &length, char * const &data) : m_Son(), 
				m_length(length), m_count() {
			m_data = new char[m_length.getLength()];				
			memcpy(m_data, data, m_length.getLength());
	}

Node::Node(Node const &node) : m_length(node.m_length) {
		int i;
		for (i=1;i>=0;i--) { 
			m_count[i] = node.m_count[i];
			m_Son[i] = node.m_Son[i];
		}
		m_data = new char[m_length.getLength()];			
		memcpy(m_data, node.m_data, m_length.getLength());
	}

Node * Node::copySubTree() const {
		Node * node = new Node(*this);
		int i;
		for (i=1;i>=0;i--) { 
			if (node->m_Son[i]) { node->m_Son[i] = m_Son[i]->copySubTree(); }
		}
		return node;
	}

Node::Node(wxString const & string) : m_Son(), 
				m_count(), m_length() { 
		size_t length = string.length()*sizeof(wxChar);
		m_length.setLength(length);
		m_data = new char[length];
		memcpy(m_data,(char *) string.c_str(),length);
		m_length.setRightOffset(1);
		m_count[(m_data[length-1]>>7)&1] = 1;
	}

Node::Node(wxChar const * const &array, size_t const &start, 
									size_t const &end, int const &count): m_Son(), m_count(), m_length()  {
		size_t length = (end-start)*sizeof(wxChar);
		m_length.setLength(length);
		m_data = new char[length];
		memcpy(m_data,(char *) (array+start),length);
		m_length.setRightOffset(1);
		m_count[(m_data[length-1]>>7)&1] = count;
	}

Node::Node(wxFile &fileIn) {
		assert(sizeof(Node) == fileIn.Read((void *) this, sizeof(Node)));
		m_data = new char[m_length.getLength()];
		assert(m_length.getLength() == fileIn.Read((void *) m_data, m_length.getLength()));
	}

Node::~Node() { // be aware that the destructor is recursive
		delete [] m_data;
		if (m_Son[0]) delete m_Son[0];
		if (m_Son[1]) delete m_Son[1];
	}

void Node::cropLeft(size_t pos) {
		assert( pos >= m_length.getLeftOffset());
		if (pos >= m_length.getBitLength()) return;
		if (pos == m_length.getLeftOffset()) {
			m_length.setLength(0);
		} else {
			m_length.setLength((pos+7)>>3);
			m_length.setRightOffset(7 - ((pos+7)&mask7));
		}
		int i = (m_data[pos>>3]>>(pos&mask7))&1;
		m_count[i] = m_count[0]+m_count[1];
		m_count[1-i] = 0;
		char *data;
		if (m_length.getLength()) {
			data = new char[m_length.getLength()];
			memcpy(data, m_data, m_length.getLength());
		} else data = 0;
		delete [] m_data;
		m_data = data;
	}

void Node::cropRight(size_t pos) { 
		if (pos <= m_length.getLeftOffset()) return;
		assert(pos <= m_length.getBitLength()); 
		if (pos == m_length.getBitLength()) { 
			m_length.setLength(0); 
			m_length.setLeftOffset(pos&mask7);	
		} else {
			m_length.setLength(m_length.getLength() - (pos>>3));
			m_length.setLeftOffset(pos&mask7);	
		}
		char *data;
		if (m_length.getLength()) {
			data	= new char[m_length.getLength()];
			memcpy(data, m_data + (pos>>3), m_length.getLength());
		} else data = 0;
		delete [] m_data;
		m_data = data;
	}

void Node::split(size_t pos) { 
		if (m_length.getLength() == 0) return;
		if (pos < m_length.getLeftOffset()) return;
		if (pos >= m_length.getBitLength()) return;
		Node * node = new Node(*this);
		int i;
		m_Son[i=getBit(pos)] = node;
		m_Son[1-i] = 0;
		node->cropRight(pos+1);
		cropLeft(pos);
	} 

size_t Node::bitsCompare(Node const &node) {
		if ((m_length.getLength() == 0)||(node.m_length.getLength()==0)) {
			return m_length.getLeftOffset();
		}
		size_t start = m_length.getLeftOffset();
		assert(start == node.m_length.getLeftOffset());
		size_t end = node.m_length.getBitLength();
		if (m_length.getBitLength() < end) end = m_length.getBitLength();
		char * &a1 = m_data;
		char * const &a2 = node.m_data;
	  while ( (start<end)&&(start<8)
										&& (((a1[start>>3]>>(start&mask7))&1)==((a2[start>>3]>>(start&mask7))&1))
					) start++;
	  while ( ((start+7)<end)
										&& (a1[start>>3]==a2[start>>3])
					) start += 8;
	  while ( (start<end)
										&& (((a1[start>>3]>>(start&mask7))&1)==((a2[start>>3]>>(start&mask7))&1))
					) start++;
		return start;
	}

void Node::add(Node * node) {
	if (node == 0) return;
	size_t pos = bitsCompare(*node);
	// cout << "pos = " << pos << endl;
	split(pos); node->split(pos);
	int i;
	for (i=1;i>=0;i--) {
		m_count[i] += node->m_count[i];
		if (m_Son[i]) { m_Son[i]->add(node->m_Son[i]); } 
		else if (node->m_Son[i]) { m_Son[i] = node->m_Son[i]->copySubTree(); }
	}
}

int Node::draw() {
	if (unif_rand() < (double(m_count[0])/double(m_count[0]+m_count[1]))) {
		return 0;
	}	else {
		return 1;
	}
}

int Node::getBit(size_t pos) { 
	assert(pos < m_length.getBitLength());
	return (m_data[pos>>3]>>(pos&mask7))&1;
}

void Node::print(wxTextOutputStream & textOut) {
	textOut << m_count[0] << _("/") << m_count[1];
}


void Node::write(wxFile &fileOut) { 
	fileOut.Write((void *) this, sizeof(Node));
	fileOut.Write((void *) m_data, m_length.getLength());
} 

size_t TreeNode::last_of(wxChar const * const &array, size_t pos, size_t const &start, 
									wxString const & pattern) {
		int length = pattern.length();
		int i;
		wxChar c;
		while(pos > start) {
			c = array[--pos];
			for (i=0;i<length;i++) {
				if (c == pattern[i]) return ++pos; 
			}
		}
		return start;
	}

size_t TreeNode::last_not_of(wxChar const * const &array, size_t pos, 
									size_t const &start, 
									wxString const & pattern) {
		int length = pattern.length();
		int i;
		wxChar c;
		bool flag;
		while(pos>start) {
			c = array[--pos];
			flag = true;
			for (i=0;i<length;i++) {
				if (c == pattern[i]) flag = false; 
			}
			if (flag) return ++pos;
		}
		return start;
	}

size_t TreeNode::first_of(wxChar const * const &array, size_t pos, size_t const &end, 
									wxString const & pattern) {
		int length = pattern.length();
		int i;
		wxChar c;
		for(;pos<end;pos++) {
			c = array[pos];
			for (i=0;i<length;i++) {
				if (c == pattern[i]) return pos; 
			}
		}
		return end;
	}

size_t TreeNode::first_not_of(wxChar const * const &array, size_t pos, 
									size_t const &end, 
									wxString const & pattern) {
		int length = pattern.length();
		int i;
		wxChar c;
		bool flag;
		for(;pos<end;pos++) {
			c = array[pos];
			flag = true;
			for (i=0;i<length;i++) {
				if (c == pattern[i]) flag = false; 
			}
			if (flag) return pos;
		}
		return end;
	}

/*
 * Selects the shortest context seen less than thresh times.
 */

wxString TreeNode::contLagSample(wxString const & starter, unsigned int thresh) {
		int i;
		wxChar * array = (wxChar *) m_buff;
		size_t ahead;
		size_t start = 0; 
		size_t end = starter.length();
		memcpy(m_buff, (char *) starter.c_str(), end*sizeof(wxChar));
		size_t sEnd = end;
		// size_t sEnd = sample(array, 0, length, BUFFLENGTH);		
		while(start < end) { 
			ahead = nextTok(array, start, sEnd, whiteSpace);
			if ((end <= ahead) 
						|| (search(array, ahead, end) > thresh)
					) {
				sEnd = sample(array, start, end, BUFFLENGTH);
				end = nextTok(array, end, sEnd, whiteSpace);
			}
			while ( (end < sEnd) 
								&& (search(array, ahead, end) > thresh)
						) {
				end = nextTok(array, end, sEnd, whiteSpace);
			}
			start = nextTok(array, start, sEnd, whiteSpace);
		}
		wxString string;
		memcpy((char *) string.GetWriteBuf(sEnd),
										m_buff, sEnd*sizeof(wxChar));
		string.UngetWriteBuf();
		return string;
	}

void TreeNode::readText(wxString const &fileName, int const &rank) {
		int i;
		wxFFile textFile(fileName, _("r"));
		wxString string;
		textFile.ReadAll(&string);
		string.Replace(_("\n"),_(" ")); 
		size_t pos;
		string.Replace(_("."),_(" . ")); // isolates . to be interpreted as a token
		string.Replace(_("?"),_(" ? ")); 
		string.Replace(_("!"),_(" ! ")); 
		string.Replace(_("'"),_(" ' ")); 
		string.Replace(_(","),_(" , ")); 
		string.Replace(_("*"),_(" * ")); 
		string.Replace(_("  "),_(" ")); 
  	wxStringTokenizer tkz(string,_(".?!"), wxTOKEN_RET_DELIMS);
		while (tkz.HasMoreTokens()) {
	    wxString token = tkz.GetNextToken(); // puts token on the stack
			// token.Trim(false);
			wxStringTokenizer tkz2(token, whiteSpace, wxTOKEN_STRTOK);
			wxString output = _("\t");
			while (tkz2.HasMoreTokens()) {
				output += tkz2.GetNextToken();
				output += _(" ");
			}
			// output.Shrink();
			wxChar const * array = output.c_str();
			size_t start, end;
			size_t tEnd = output.length(); 
			// start = first_not_of(array, 0, tEnd, whiteSpace); 
			start = 0; 
			if (rank) { 
				end = first_not_of(array, 0, tEnd, whiteSpace); 
				for (i = rank; i; i--) { 
					end = nextTok(array, end, tEnd, whiteSpace);
				}
			} else { end = tEnd; }
			{
				Node node(array, start, end);
				if (m_root) m_root->add(&node);
				else m_root = new Node(node);
				start = first_not_of(array, 0, tEnd, whiteSpace); 
			}
			while(start<end) {
				Node node(array, start, end);
				m_root->add(&node);
				start = nextTok(array, start, tEnd, whiteSpace);
				end = nextTok(array, end, tEnd, whiteSpace);
			}

			/*
  		wxStringTokenizer tkz2(token,_(" "), wxTOKEN_STRTOK);
			while (tkz2.HasMoreTokens()) {
				pos = tkz2.GetPosition();
				wxString token2 = token.Mid(pos);
				if (rank) cropString(token2, 0, rank);
				Node node(token2.Trim(false));
				if (m_root) m_root->add(&node);
				else m_root = new Node(node);
	   		tkz2.GetNextToken(); 
			}
			*/

		}
	}

void TreeNode::readTextFB(wxString const &fileName, wxString const & starter, int weight) { 
		int i;
		wxFFile textFile(fileName, _("r"));
		wxString string;
		textFile.ReadAll(&string);
		string.Replace(_("\n"),_(" ")); 
		size_t pos;
		string.Replace(_("."),_(" . ")); // isolates . to be interpreted as a token
		string.Replace(_("?"),_(" ? ")); 
		string.Replace(_("!"),_(" ! ")); 
		string.Replace(_("'"),_(" ' ")); 
		string.Replace(_(","),_(" , ")); 
		string.Replace(_("*"),_(" * ")); 
		string.Replace(_("("),_(" ( ")); 
		string.Replace(_(")"),_(" ) ")); 
		string.Replace(_("  "),_(" ")); 
		string.Replace(_("["),_("{")); 
		string.Replace(_("]"),_("}")); 
  	wxStringTokenizer tkz(string,_(".?!"), wxTOKEN_RET_DELIMS);
		while (tkz.HasMoreTokens()) {
	    wxString token = tkz.GetNextToken(); // puts token on the stack
			// token.Trim(false);
			wxStringTokenizer tkz2(token, whiteSpace, wxTOKEN_STRTOK);
			wxString output = starter;
			while (tkz2.HasMoreTokens()) {
				output += tkz2.GetNextToken();
				output += _(" ");
			}
			size_t length = output.length();
			if (length == starter.length()) break;
			output = output + output;
			wxChar const * array = output.c_str();
			size_t start, end, sEnd;
			size_t tEnd = output.length(); 
			// start = first_not_of(array, 0, tEnd, whiteSpace); 
			start = 0; end = length; 
			while(start < length) {
				/*
				sEnd = first_of(array, start, end, starter);
				if (sEnd >= end) {
				cout << "Sentence without beginning : \n" << 
								output.mb_str(wxConvUTF8) << endl;
				}
				*/
				Node node(array, start, end, weight);
				if (m_root) m_root->add(&node);
				else m_root = new Node(node);
				start = nextTok(array, start, length);
				end = start + length; 
			}
		}
	}

int TreeNode::remove_all(wxString const &string, int count) { 
	static int length;
	static int searchCount;
	length = string.length();
	memcpy(m_buff, (void *)string.c_str(), length*sizeof(wxChar));	
	if ((searchCount = search_g((wxChar *)m_buff, 0, length)) < count) { 
		count = searchCount;
	}
	if (count) {
		remove_all((wxChar *)m_buff, 0, length, count);
	}
	return count;
} 

void TreeNode::add_all(wxString const &string, int count) { 
	static int length;
	if (count == 0) return;
	length = string.length();
	memcpy(m_buff, (void *)string.c_str(), length*sizeof(wxChar));
	add_all((wxChar*) m_buff, 0, length, count);
}

void TreeNode::add(wxChar *array, size_t start, size_t end, 
								int const &count) {
	if (count <= 0) return;
	if (m_root) {
		Node node(array, start, end, count);
		m_root->add(&node);
	} else {
		m_root = new Node(array, start, end, count);
	}
	/*
	if (m_root == 0) m_root = new Node();
	Node node(array, start, end, count);
	m_root->add(&node);
	*/
}

void TreeNode::add_all(wxChar *array, size_t start, size_t end, 
								int const &count) {
	if (count <= 0) return;
	size_t mid = start;
	size_t length = end - start;
	rollover(array, mid, end, end); 
	end = mid;
	if (m_root == 0) m_root = new Node();
	while (start < mid) { 
		Node node(array, start, end, count);
		m_root->add(&node);
		start = nextTok(array, start, mid);
		end = start + length;
	}
}

void TreeNode::remove_all(wxChar *array, size_t start, size_t end, int count) { 
	if (count <= 0) return;
	size_t mid = start;
	size_t length = end - start; 
	rollover(array, mid, end, end);
	end = mid;
	while (start < mid) { 
		remove(array, start, end, count);
		start = nextTok(array, start, mid);
		end = start + length;
	} 
} 

void TreeNode::remove_all(wxChar *array, size_t start, size_t end) { 
	int count = search_g(array, start, end);	
	if (count <= 0) return;
	size_t mid = start;
	size_t length = end - start; 
	rollover(array, mid, end, end);
	end = mid;
	while (start < mid) { 
		remove(array, start, end, count);
		start = nextTok(array, start, mid);
		end = start + length;
	} 
} 

void TreeNode::remove(wxChar * array, size_t a_start, size_t end, int a_count) {
		int i;
		int v_count;
		size_t newEnd;
		size_t start = a_start;
		if (start >= end) {
			m_pos = m_root;
			m_offset = 0;
			return;
		}
		start *= sizeof(wxChar);
		end *= sizeof(wxChar);
		char * a = (char *) array;
		start <<= 3; 
		end <<= 3; 
		m_pos = m_root;
		m_offset = 0;
		size_t bitLength;
		char * b = m_pos->m_data;
		while(start < end) {
			if (m_pos->m_length.getLength()) { 
				bitLength = m_pos->m_length.getBitLength();
				while (
								(start<end)&&(m_offset<bitLength)&&((start&mask7)!=0)
								&& (((a[start>>3]>>(start&mask7))&1)
								 	==((b[m_offset>>3]>>(m_offset&mask7))&1))
							) { start++; m_offset++; }
				while (
								((start+7)<end)&&((m_offset+7)<bitLength)
								&& (a[start>>3]==b[m_offset>>3])
							) { start += 8; m_offset += 8; }
				while (
								(start<end)&&(m_offset<bitLength)
								&& (((a[start>>3]>>(start&mask7))&1)
								 	==((b[m_offset>>3]>>(m_offset&mask7))&1))
							) { start++; m_offset++; }
				if (m_offset != bitLength) break;
			} 
			if 	(start<end)  {
				v_count = (m_pos->m_count[i=((a[start>>3]>>(start&mask7))&1)] 
												-= a_count); 
				if (v_count < 0) { 
					cerr << "v_count = " << v_count << endl;
					cerr << wxString(array+a_start, start/(sizeof(wxChar)<<3)-a_start).mb_str(wxConvUTF8) << "|\n";
					assert(false);
				} 
				start++; 
				m_pos = m_pos->m_Son[i]; 
			} else break;
			if (m_pos) {
				m_offset = m_pos->m_length.getLeftOffset();
				b = m_pos->m_data;
			} else break;
		}
		if (start < end) {
			cerr << "Warning : start >= end,\n"; 
			cerr << "start = " << start << "\nend = " << end << endl;
			assert(false);
		}

	}


int TreeNode::search_g(wxChar * array, size_t start, size_t end) {
		int i;
		int count;
		size_t newEnd;
		if (start >= end) {
			m_pos = m_root;
			m_offset = 0;
			return 0;
		}
		start *= sizeof(wxChar);
		end *= sizeof(wxChar);
		char * a = (char *) array;
		start <<= 3; 
		end <<= 3; 
		m_pos = m_root;
		m_offset = 0;
		size_t bitLength;
		char * b = m_pos->m_data;
		while(start < end) {
			if (m_pos->m_length.getLength()) { 
				bitLength = m_pos->m_length.getBitLength();
				while (
								(start<end)&&(m_offset<bitLength)&&((start&mask7)!=0)
								&& (((a[start>>3]>>(start&mask7))&1)
								 	==((b[m_offset>>3]>>(m_offset&mask7))&1))
							) { start++; m_offset++; }
				while (
								((start+7)<end)&&((m_offset+7)<bitLength)
								&& (a[start>>3]==b[m_offset>>3])
							) { start += 8; m_offset += 8; }
				while (
								(start<end)&&(m_offset<bitLength)
								&& (((a[start>>3]>>(start&mask7))&1)
								 	==((b[m_offset>>3]>>(m_offset&mask7))&1))
							) { start++; m_offset++; }
				if (m_offset != bitLength) break;
			} 
			if 	( (start<end) && 
						(count = m_pos->m_count[i=((a[start>>3]>>(start&mask7))&1)]) 
					) { 
				start++; 
				m_pos = m_pos->m_Son[i]; 
			} else break;
			if (m_pos) {
				m_offset = m_pos->m_length.getLeftOffset();
				b = m_pos->m_data;
			} else break;
		}
		if (start < end) { m_pos = 0; return 0; }
		else if (m_pos) { 
			return count - m_pos->m_count[0] - m_pos->m_count[1];
		} else {
			return count;
		}	
	}

void copyHist( wxChar * &hist, wxChar * const &array, size_t start, 
								size_t sEnd, size_t end) {
		assert(start <= sEnd);
		assert(sEnd < end);
		/*
		if (sEnd >= end) {
			cout << "sEnd = " << sEnd << endl;
			cout << "end = " << end << endl;
			wxString output(array+start, end-start);
			cout << output.mb_str(wxConvUTF8) << endl;
		}
		*/
		memcpy((void *) hist, (const void *) (array + sEnd), (end - sEnd)*sizeof(wxChar));
		hist += end - sEnd;
		*hist = '^'; hist++;
		*hist = ' '; hist++;
		memcpy((void *) hist, (const void *) (array + start), (sEnd - start)*sizeof(wxChar));
		hist += sEnd - start;
}

wxString TreeNode::FBsample(wxString const & starter, int gen_cycles, int read_thresh, int fgen_thresh, int bgen_thresh, bool global, wxString & hist) {
	wxChar * array = (wxChar *) m_buff;
	wxString outputString;
	wxChar * outputStart = outputString.GetWriteBuf(BUFFLENGTH);
	wxChar * output = outputStart;
	wxChar *histBuf = hist.GetWriteBuf(BUFFLENGTH);
	wxChar *histPtr = histBuf;
	size_t start, end, sEnd;
  size_t pos, cond;	
	size_t length = starter.length();
	int i, j;
	memcpy((void *) (array), (const void *) starter.c_str(), length*sizeof(wxChar));
	start = 0;
	pos = length;
	end = sample(array, start, pos);
	rollover(array, start, pos, end);
	sEnd = first_of(array, start, end, starter);
	copyHist(histPtr, array, start, sEnd, end);
	for (j = MAX_SENTENCE_LENGTH;j;j--) {   
		pos = nextTok(array, start, sEnd);
		while ((pos < sEnd )
						&&(search(array, start, pos)>read_thresh)) {
			pos = nextTok(array, pos, sEnd);
		}
		memcpy((void *) output, (const void *) (array + start), (pos - start)*
									 sizeof(wxChar));	
		output += (pos - start);
		rollover(array, start, pos, end);
		sEnd = first_of(array, start, end, starter);
		copyHist(histPtr, array, start, sEnd, end);
		if (nextTok(array, start, sEnd) >= sEnd) {
			memcpy((void *) output, (const void *) (array + start), (sEnd - start)*
									 sizeof(wxChar));	
			output += (sEnd - start);
			break;
		}

		for (i= gen_cycles; i; i--) {
			for (cond = (global) ? sEnd : start; 
											(cond < end)&&(search(array, start, cond) > bgen_thresh)
											; cond = nextTok(array, cond, end)) {} 	
			assert(search(array, start, cond));
			end = sample(array, start, cond);
			sEnd = first_of(array, start, end, starter);
			copyHist(histPtr, array, start, sEnd, end);
			for (cond = (global) ? sEnd : end; 
					(cond > start)&&(search(array, cond, end) > fgen_thresh); 
					cond = prevTok(array, cond, start)) {}
			assert(search(array, cond, end));
			pos = end;
			end = sample(array, cond, end);
			start = cond;
			rollover(array, start, pos, end);
			sEnd = first_of(array, start, end, starter);
			copyHist(histPtr, array, start, sEnd, end);
		}
	}

	*histPtr = 0;
	hist.UngetWriteBuf(histPtr-histBuf);
	*output = 0;
	outputString.UngetWriteBuf(output - outputStart);
	return outputString;
}

void TreeNode::gatherNew (wxString const &header, TreeNode &oldExp,
	TreeNode &newExp ) {
	size_t pos;
	memcpy(m_buff, (void *)header.c_str(), pos = header.length()*sizeof(wxChar));
	if (m_Count = search(header)) {
					gatherNew( m_pos, pos << 3, oldExp, newExp);
					// gatherNew( m_pos, 0, oldExp, newExp);
	}
}	

void TreeNode::gatherNew(Node * node, size_t pos, TreeNode &oldExp,
							TreeNode &newExp) 	{ 
	int i, j;
	size_t end;
	// static int count = 0;
	size_t offset = m_offset;
	m_offset = 0;
	wxChar *data = (wxChar *) m_buff;
	if (node == 0) {
		end  = (pos>>3)/sizeof(wxChar);
		// cerr << wxString(data,end).mb_str(wxConvUTF8) << "|\n";
		if ((!oldExp.search(data,0,end))&&(!newExp.search(data,0,end))) {
			newExp.add(data,0,end);
		}
		return;
	}	
	if ((node->m_count[0] + node->m_count[1])< m_Count) {
		end = (pos>>3)/sizeof(wxChar);
		if ((!oldExp.search(data,0,end))&&(!newExp.search(data,0,end))) {
			newExp.add(data,0,end);
		}
	}
	if (node->m_length.getLength()) {
		size_t bitLength;
		memcpy(m_buff+(pos>>3), node->m_data + (offset>>3), node->m_length.getLength() - (offset >> 3));
		bitLength = node->m_length.getBitLength();
		pos = (pos&mask2) + bitLength - offset;
	}
	i = node->draw();
	for (j=2;j;j--,i = 1-i) {
		if (i) m_buff[pos>>3] |= (1<<(pos&mask7));
		else m_buff[pos>>3] &= ~(1<<(pos&mask7));
		if (m_Count = node->m_count[i]) {
			gatherNew(node->m_Son[i], pos+1, oldExp, newExp);
		}
	}

	return;
}

bool Label::remap() {
	m_end = 0;
	m_tree->setCount();
	remap(m_tree->m_root, 0, 0);
	if (m_end) return true;
	else return false;
}	

void Label::remap (Node * node, size_t pos, size_t charPos)
	{ 
	int i, j;
	size_t newCharPos;
	char *buff = (char *) m_stem;
	wxChar *charBuff = (wxChar *) TreeNode::m_buff;
	wxString rightLabel, leftLabel;

	if ((node == 0)||((node->m_count[0] + node->m_count[1])< m_tree->m_Count)) {
		newCharPos = TreeNode::nextTok(m_stem,0,charPos);
		if ((m_stem[0]!='[')&&(m_stem[0]!=']')) {
			cerr << "rightLabel = " 
							<< wxString(m_stem, newCharPos).mb_str(wxConvUTF8) << "|\n";
			assert(false);
		}
		rightLabel = wxString(m_stem+1, newCharPos - 2);
		// cerr << "remap rightLabel = " << rightLabel.mb_str(wxConvUTF8) << "|\n"; 
		m_end = charPos - newCharPos;
		memcpy(TreeNode::m_buff,(char *) (m_stem+newCharPos),
										m_end*sizeof(wxChar));
		charBuff[m_end] = m_stem[0];
		m_end++;
		newCharPos = m_grammar->sample_p(charBuff,0,m_end,_(" "),rightLabel);
		if (newCharPos > m_end) { 
			leftLabel = wxString(charBuff+m_end,newCharPos - m_end - 1);
			// cerr << "remap : second label found !\n";
			// coucou
			m_tree->merge_all(rightLabel, leftLabel);	
			// cerr << "rightLabel = " << rightLabel.mb_str(wxConvUTF8) << "|\n";
			// cerr << "leftLabel = " << 	leftLabel.mb_str(wxConvUTF8) << "|\n";
			if (m_tree != m_grammar) {
					m_grammar->merge_all(rightLabel, leftLabel);
			}
			m_end = charPos; 
			return;
		} else {
			m_end = 0;
		}
	}

	if (node == 0) { 
		return;
	}

	if (node->m_length.getLength()) {
		size_t bitLength;
		memcpy(buff+(pos>>3), node->m_data, node->m_length.getLength());
		bitLength = node->m_length.getBitLength();
		pos = (pos&mask2) + bitLength;
	}
	i = node->draw();
	newCharPos = ((pos+1)>>3)/sizeof(wxChar);
	for (j=2;j;j--,i = 1-i) {
		if (i) buff[pos>>3] |= (1<<(pos&mask7));
		else buff[pos>>3] &= ~(1<<(pos&mask7));
		if ((m_tree->m_Count = node->m_count[i])&&
					((charPos > 0)||(newCharPos == 0)||
					(*m_stem == '[')||(*m_stem  == ']'))&&
					((charPos > 1)||(newCharPos < 2)||
						 	(*m_stem != '[')||(m_stem[1] != '0'))
			 )
		{
			remap(node->m_Son[i], pos+1, newCharPos);
			if (m_end) return;
		}
	}

	return;
}

wxString TreeNode::read(wxChar const *a, size_t start, size_t end) { 
	wxString output = _("");
	if (m_Count = search(a, start, end)) read(m_pos, 0, output);
	return output;
}

wxString TreeNode::read(wxString starter) { 
	wxString output = _("");
	if (m_Count = search(starter)) read(m_pos, 0, output);
	return output;
} 

void TreeNode::read(Node *node, size_t pos, wxString &output) { 
	int i, j;
	// static int count = 0;
	size_t offset = m_offset;
	m_offset = 0;
	wxChar *data = (wxChar *) m_buff;
	if (node == 0) {
		for (i=0;i<((pos>>3)/sizeof(wxChar));i++) {
			output << data[i];
		}
		output << _("\n");
		// count = 0;
		return;
	}	
	if ((node->m_count[0] + node->m_count[1])< m_Count) {
		for (i=0; i < ((pos>>3)/sizeof(wxChar));i++) {
			output << data[i];
		}
		output << _("\n");
	}
	if (node->m_length.getLength()) {
		size_t bitLength;
		memcpy(m_buff+(pos>>3), node->m_data + (offset>>3), node->m_length.getLength() - (offset >> 3));
		bitLength = node->m_length.getBitLength();
		pos = (pos&mask2) + bitLength - offset;
	}
	i = node->draw();
	for (j=2;j;j--,i = 1-i) {
		if (i) m_buff[pos>>3] |= (1<<(pos&mask7));
		else m_buff[pos>>3] &= ~(1<<(pos&mask7));
		// pos++; error ! increments pos twice !
		if (m_Count = node->m_count[i]) {
			read(node->m_Son[i], pos+1, output);
		}
	}

	return;
}

size_t TreeNode::insert(wxChar *array, size_t const &dStart, size_t 
								const &dEnd, size_t const &sEnd, size_t const &iStart, size_t 
								const &iEnd) {
	if ((dEnd < dStart)||(sEnd < dEnd)||(iStart<sEnd)||(iEnd<iStart)) {
					return sEnd; }
	memcpy((void*) (array+iEnd), (void *) (array + iStart), 
									(iEnd-iStart)*sizeof(wxChar));
	memcpy((void*) (array+2*iEnd - iStart), (void *) (array + dEnd), 
									(sEnd-dEnd)*sizeof(wxChar));
	memcpy((void*) (array+dStart), (void *)(array+iEnd), 
									(iEnd-iStart+sEnd-dEnd)*sizeof(wxChar));
	return dStart + iEnd - iStart + sEnd - dEnd;
}

bool TreeNode::mergeGrammar() {
	wxChar *array = (wxChar *) m_buff;
	size_t start_a, end_a, end_b;
	wxString string;
	bool ok = true;
	bool succeeded = true;
	array[0] = ']'; // we are looking for a label of the form ]_i
	int count;

	while(ok) {
		// looking for expression ]_i a
 		if ((end_a = sample(array,0,1)) == 1) { // there is no such label
			ok = false;
			break;
		}

// debug check, to be eventually removed
		if ((count = search_g(array, 0, end_a)) < 1) {
			cerr << "count = " << count << endl;
			cerr << wxString(array, end_a).mb_str(wxConvUTF8) << "|\n";
			printS(_("grammarDebug.txt"));
			assert(false);
		}
		// removes all permutations of expression ]_i a 
		remove_all(array,0,end_a,1); 
		start_a = nextTok(array,0,end_a);
		array[0] = '['; // we now have [_i a
		// looking for an expression [_i b starting with [_i
		end_b = sample(array,0,start_a,end_a); // copy it at the end of [_i a 
		if (end_b == end_a) {
			succeeded = false;
			continue;
		}
		// add all permutations of ab
		add_all(array, start_a, end_b, 1); 
		// add [_i at the end of string b
		memcpy((void *)(array + end_b),(void *) array, start_a*sizeof(wxChar));
		array[0] = ']'; // not needed ?
		if ((count = search_g(array, end_a, end_b+start_a)) < 1) {
			cerr << "count = " << count << endl;
			cerr << wxString(array, end_a).mb_str(wxConvUTF8) << "|\n";
			cerr << wxString(array+end_a, end_b+start_a - end_a).mb_str(wxConvUTF8) << "|\n";
			printS(_("grammarDebug.txt"));
			assert(false);
		}
		// removes expression b [_i and its permutations
		remove_all(array,end_a,end_b+start_a,1); 
		/*

		if ((count = search_g(array, 0, end_a)) < 1) {
			cerr << "count = " << count << endl;
			cerr << wxString(array, end_a).mb_str(wxConvUTF8) << "|\n";
			cerr << wxString(array+end_a, end_b+start_a - end_a).mb_str(wxConvUTF8) << "|\n";
			printS(_("grammarDebug.txt"));
			assert(false);
		}
		remove_all(array,0,end_a,1);

		*/
	}
	return succeeded; 
	// succeeded = false when we erased 
	// some expression of the form [_i a ]_i b[_i a ]_i b
}

// requires that the initial symbol is [0
wxString TreeNode::sampleGrammar(wxString &hist) {
	wxChar *array = (wxChar *) m_buff;
	wxChar histArray[BUFFLENGTH];
	size_t histInd = 0;
	wxString starter(_("[0 ]0 "));
	size_t sEnd = starter.length();
	size_t start, end, iStart, iEnd;
	memcpy((void *) array, (void *) starter.c_str(), 
		sEnd*sizeof(wxChar)); 
	start = nextTok(array, 0, sEnd);
	end = nextTok(array, start, sEnd);
	while(start < end) {
		if (array[start] == ']') { 
			memcpy((void *)(array+sEnd), (void *)(array+start), 
											(end-start)*sizeof(wxChar));
			array[sEnd] = '[';
			iEnd = sample(array, sEnd, iStart = sEnd + end - start);

			if (iEnd > iStart) {
				memcpy((void *)(histArray+histInd),(void *)(array+sEnd),
											(iEnd-sEnd)*sizeof(wxChar));
				histInd += iEnd-sEnd;
				histArray[histInd++] = '\n';

				sEnd = insert(array, start, end, sEnd, iStart, iEnd);	
				end = nextTok(array, start, sEnd);
			} else {
				start = end;
				end = nextTok(array, start, sEnd); 
			}
		} else {
			start = end;
			end = nextTok(array, start, sEnd); 
		}
	}
	hist = wxString(histArray, histInd);
	return wxString(array, sEnd);
}

TreeNode::TreeNode(TreeNode const &tree): Tree<Node>(), 
				m_labelCount(tree.m_labelCount), 
		m_labelMax(tree.m_labelMax) {
	m_Count = tree.m_Count;
	m_CrIndex = tree.m_CrIndex;
	m_root = tree.m_root->copySubTree();
	m_pos = m_root;
}

TreeNode& TreeNode::operator=(TreeNode const &rightTree) {
	if (this != &rightTree) {
		delete m_root;
		m_root = rightTree.m_root->copySubTree(); 
		m_pos = m_root;
		m_labelMax = rightTree.m_labelMax;
		m_labelCount = rightTree.m_labelCount;
		m_Count = rightTree.m_Count;
	}
	return *this;
}

void Label::write(wxString const &filename) {
	wxFile fileOut(filename, wxFile::write);
	fileOut.Write((void *) this, sizeof(Label));
}

Label::Label(TreeNode *tree, wxString const &filename) { 
	wxFile fileIn(filename, wxFile::read); 
	fileIn.Read((void *)this, sizeof(Label));
	m_stem = new wxChar[BUFFLENGTH];
	m_tree = tree;
	m_grammar = tree;
}

size_t TreeNode::nextTok(wxChar const * const &array, size_t const &pos, 
		size_t const &end, 
		wxString const & pattern) {
	return first_not_of(array, 
			first_of(array, pos, end, pattern),
			end, pattern);
}

size_t TreeNode::prevTok(wxChar const * const &array, size_t const &pos, 
		size_t const &start, 
		wxString const & pattern) {
	return last_of(array, 
			last_not_of(array, pos, start, pattern),
			start, pattern);
}


void TreeNode::add (wxString const & a_string) {
	if (m_root) {
		Node node(a_string);
		m_root->add(&node);
	} else {
		m_root = new Node(a_string);
	}
}


	/*
	 * 	 *  Selects the largest context seen at least thresh times.
	 * 	 	 */

wxString TreeNode::contextSample(wxString const & starter, unsigned int thresh) {
	int i;
	wxChar * array = (wxChar *) m_buff;
	size_t start = 0; 
	size_t end = starter.length();
	memcpy(m_buff, (char *) starter.c_str(), end*sizeof(wxChar));
	size_t sEnd = end;
	// size_t sEnd = sample(array, 0, length, BUFFLENGTH);		
	while(start < end) { 
		if ((end <= nextTok(array, start, sEnd, whiteSpace)) 
				|| (search(array, start, end) >= thresh)
			 ) {
			sEnd = sample(array, start, end, BUFFLENGTH);
			end = nextTok(array, end, sEnd, whiteSpace);
		}
		while ( (end < sEnd) 
				&& (search(array, start, end) >= thresh)
				) {
			end = nextTok(array, end, sEnd, whiteSpace);
		}
		start = nextTok(array, start, sEnd, whiteSpace);
	}

	wxString string;
	memcpy((char *) string.GetWriteBuf(sEnd),
			m_buff, sEnd*sizeof(wxChar));
	string.UngetWriteBuf();
	return string;
}


wxString TreeNode::markovSample(wxString const &starter, int rank) {
	size_t start, end;
	int i;
	wxChar * array = (wxChar *) m_buff;
	memcpy(m_buff, (char *) starter.c_str(), starter.length()*sizeof(wxChar));
	size_t sEnd = sample(array, 0, starter.length(), BUFFLENGTH);		
	// for (	start = end = first_not_of(array, 0, 
	for (	start = end = 0, i=rank+1; 
			i && (end < sEnd);
			i--) {
		end = nextTok(array, end, sEnd, whiteSpace);
	}									
	while (start < end) {
		start = nextTok(array, start, end, whiteSpace);
		sEnd = sample(array, start, end, BUFFLENGTH);		
		end = nextTok(array, end, sEnd, whiteSpace);
	}

	wxString string;
	memcpy((char *) string.GetWriteBuf(sEnd),
			m_buff, sEnd*sizeof(wxChar));
	string.UngetWriteBuf();
	return string;

}


bool TreeNode::cropString(wxString &string, int start, int end) {
	size_t pos[end];
	int i;
	wxStringTokenizer tkz(string,_(" "), wxTOKEN_STRTOK);
	for (i=0; (i<=end)&&tkz.HasMoreTokens(); i++) {
		pos[i] = tkz.GetPosition();
		tkz.GetNextToken(); 
	}
	if (i>end) {
		string = string.Mid(pos[start], pos[end] - pos[start]).Trim(false);
		return true;
	} else if (i == end) {
		string = string.Mid(pos[start]).Trim(false);
		return true;
	} else {
		return false;
	}
}


size_t TreeNode::sample(wxChar * array, size_t a_start, 
		size_t a_end, size_t a_newStart) {
	int i, total;
	size_t length, bitLength;
	char * charA = (char *) array;
	size_t end = a_end*sizeof(wxChar)*8;
	size_t a, b;
	size_t shift = 0;
	if (a_newStart < BUFFLENGTH) {
		shift = (a_newStart - a_end) * sizeof(wxChar);
	} 
	// search will set the m_pos pointer
	if (!search(array, a_start, a_end)) return a_end + shift/sizeof(wxChar);
	while(m_pos) {
		length = m_pos->m_length.getLength();
		if (length > (m_offset >> 3)) {
			bitLength = m_pos->m_length.getBitLength();
			assert(shift + (end>>3) + length - (m_offset >> 3) <= BUFFLENGTH*sizeof(wxChar));
			memcpy(charA + shift + (end>>3), m_pos->m_data + (m_offset>>3), 
					length-(m_offset>>3));
			end += bitLength - m_offset;
			m_offset = bitLength;
		}
		i = m_pos->draw();
		// if (m_pos->m_count[i] <= thresh) break;
		if (i) {
			charA[(end>>3)+shift] |= (1<<(end&mask7)); 
		} else {
			charA[(end>>3)+shift] &= ~(1<<(end&mask7));
		}
		end++;
		if (!m_pos->m_Son[i]) break;
		if (m_pos->m_count[i] > (total = m_pos->m_Son[i]->m_count[0] + 
					m_pos->m_Son[i]->m_count[1])) {
			if (unif_rand() >= ((double)total)/((double) m_pos->m_count[i])) break;
		}
		m_pos = m_pos->m_Son[i];
		// if (!m_pos) break;
		m_offset = m_pos->m_length.getLeftOffset();
	}
	return ((end>>3)+shift)/sizeof(wxChar);
}

wxString TreeNode::sample(wxString const &string, size_t pos) {
	memcpy(m_buff, (char *) string.c_str(), string.length()*sizeof(wxChar));
	wxString rString;
	size_t stringLength = sample((wxChar *)m_buff,pos,string.length(),BUFFLENGTH);
	memcpy((char *) rString.GetWriteBuf(stringLength),
			m_buff, stringLength*sizeof(wxChar));
	rString.UngetWriteBuf();
	return rString;
}


int TreeNode::search(wxChar const * array, size_t start, size_t end) {
	int i;
	int count;
	size_t newEnd;
	if (m_root == 0) return 0;
	if (start >= end) {
		m_pos = m_root;
		m_offset = 0;
		return m_root->m_count[0] + m_root->m_count[1];
	}
	start *= sizeof(wxChar);
	end *= sizeof(wxChar);
	char * a = (char *) array;
	start <<= 3; 
	end <<= 3; 
	m_pos = m_root;
	m_offset = 0;
	count = m_pos->m_count[0] + m_pos->m_count[1];
	size_t bitLength;
	char * b = m_pos->m_data;
	while(start < end) {
		if (m_pos->m_length.getLength()) { 
			bitLength = m_pos->m_length.getBitLength();
			while (
					(start<end)&&(m_offset<bitLength)&&((start&mask7)!=0)
					&& (((a[start>>3]>>(start&mask7))&1)
						==((b[m_offset>>3]>>(m_offset&mask7))&1))
					) { start++; m_offset++; }
			while (
					((start+7)<end)&&((m_offset+7)<bitLength)
					&& (a[start>>3]==b[m_offset>>3])
					) { start += 8; m_offset += 8; }
			while (
					(start<end)&&(m_offset<bitLength)
					&& (((a[start>>3]>>(start&mask7))&1)
						==((b[m_offset>>3]>>(m_offset&mask7))&1))
					) { start++; m_offset++; }
			if (m_offset != bitLength) break;
		} 
		if 	( (start<end) && 
				(count = m_pos->m_count[i=((a[start>>3]>>(start&mask7))&1)]) 
				) { 
			start++; 
			m_offset = 0;
			m_pos = m_pos->m_Son[i]; 
		} else break;
		if ((m_pos)&&(start<end)) {
			m_offset = m_pos->m_length.getLeftOffset();
			b = m_pos->m_data;
		} else break;
	}
	if (start < end) { m_pos = 0; return 0; }
	else if (m_offset) {
		return m_pos->m_count[0] + m_pos->m_count[1];
	} else {
		return count;
	}
}

int TreeNode::search(wxString const &string, size_t pos) {
	int i;
	int count;
	size_t start = 0;
	if (pos >= string.length()) return 0;
	if (m_root == 0) return 0;
	size_t end = (string.length()-pos)*sizeof(wxChar)*8;
	char * a = (char *) (string.c_str() + pos);
	m_pos = m_root;
	m_offset = 0;
	size_t bitLength;
	char * b = m_pos->m_data;
	while(start < end) {
		if (m_pos->m_length.getLength()) { 
			bitLength = m_pos->m_length.getBitLength();
			while (
					(start<end)&&(m_offset<bitLength)&&((start&mask7)!=0)
					&& (((a[start>>3]>>(start&mask7))&1)
						==((b[m_offset>>3]>>(m_offset&mask7))&1))
					) { start++; m_offset++; }
			while (
					((start+7)<end)&&((m_offset+7)<bitLength)
					&& (a[start>>3]==b[m_offset>>3])
					) { start += 8; m_offset += 8; }
			while (
					(start<end)&&(m_offset<bitLength)
					&& (((a[start>>3]>>(start&mask7))&1)
						==((b[m_offset>>3]>>(m_offset&mask7))&1))
					) { start++; m_offset++; }
			if (m_offset != bitLength) break;
		} 
		if 	( (start<end) && 
				(count = m_pos->m_count[i=((a[start>>3]>>(start&mask7))&1)]) 
				) { 
			start++; 
			m_offset = 0;
			m_pos = m_pos->m_Son[i]; 
		} else break;
		if (m_pos) {
			m_offset = m_pos->m_length.getLeftOffset();
			b = m_pos->m_data;
		} else break;
	}
	if (start < end) { m_pos = 0; return 0; }
	else return count;
}


void TreeNode::stringTree() {
				Node *node;
				m_root = new Node(_("This is old"));
				m_root->add(node = new Node(_("This was new")));
				delete node;
				m_root->add(node = new Node(_("This is good")));
				m_root->add(node);
				delete node;
				m_root->add(node = new Node(_("This was good")));
				delete node;
				m_root->add(node = new Node(_("Another one")));
				delete node;
}

/*
 * Now comes the low level recursive read. We test the pointers
 * and restore them when they are not zero. Maybe we can start
 * with recursive tree manipulation functions, after all. 
 * They are so simple to write ! I tried by best to give 
 * them a small footprint, passing arguments by const reference
 * when possible.
 */
template<typename TNode>
void Tree<TNode>::read(TNode * &node, wxFile &fileIn) {
	node = new TNode(fileIn); 
	if (node->m_Son[0]) {
		read(node->m_Son[0], fileIn);
	}
	if (node->m_Son[1]) {
		read(node->m_Son[1], fileIn);
	}
}

	  /* 
		 * And the internal recursive write
		 * function. Very simple : just writes the
		 * nodes in preorder. Passing const references
		 * is supposed to aid the compiler into optimizing
		 * function calls. 
		 */
template<typename TNode>
void Tree<TNode>::write(TNode * const &node, wxFile &fileOut) {
	node->write(fileOut);
	if (node->m_Son[0]) {
		write(node->m_Son[0],fileOut);
	}
	if (node->m_Son[1]) {
		write(node->m_Son[1],fileOut);
	}
}


/*
 * The tree pretty printer. Useful for visual inspection and 
 * debugging.
 */
template<typename TNode>
void Tree<TNode>::print(wxTextOutputStream & textOut,
		TNode * const &node, int depth) {
	int i,j;
	if (m_CrIndex) {
		m_CrIndex = 0;
		textOut << endl;
		for (j=depth;j>0;j--) { textOut << _("|\t"); }
	}
	textOut << m_Count++ << _(":");
	node->print(textOut);
	for (i=0;i<=1;i++) {
		if (node->m_Son[i]) {
			textOut << _("\t");
			print(textOut, (TNode * const &)(node->m_Son[i]), depth+1);
		} else { m_CrIndex++; }
	}
}

template<typename TNode>
void Tree<TNode>::print(wxTextOutputStream & textOut) {
				  m_Count = 1;
					  m_CrIndex = 1;
						  print(textOut, m_root, 0);
							  textOut << _("\n\n");
}

template class Tree<Node>;

